(*** DATABASE MANAGEMENT FUNCTIONS **)

(* HELPER FUNCTIONS *)
let rec substituteTable tablelist table acc = match tablelist with
| [] -> acc
| t::ts -> if (t.tablename = table.tablename) then
             substituteTable ts table (table :: acc)
           else
             substituteTable ts table (t :: acc);;

let rec getTableAux tablelist tablename = match tablelist with
| [] -> None
| t::ts -> if (t.tablename = tablename) then
             Some t
           else
             getTableAux ts tablename;;
             
let rec getTable tablelist tablename = 
    let t = getTableAux tablelist tablename in
    match t with 
    | None -> failwith ("table not found:" ^ tablename)
    | Some table -> table;;


let substituteRowDatalist (table : mltable) newrowdatalist =
{
  table with rowdatalist = newrowdatalist
};;

let rec length lst = match lst with
| [] -> 0
| x::xs -> 1 + (length xs);;


let int_of_coldatavalue v = match v with
| Int n -> n
| _ -> 0;;

let padStringRight s = 
      Printf.sprintf "%-25s" s ;;

(* FILTER/PREDICATE FUN *)
let compareData d1 d2 op = match op with
| OpEq -> d1 = d2
| OpNe -> d1 <> d2;;

let rec evalPred p rowdata = match rowdata with
| [] -> false   (* No matching columns in row *)
| d::ds -> if (d.datacolname = p.preddata.datacolname) then
             compareData d.value p.preddata.value p.predoper
           else
             evalPred p ds;;

let rec filterrowdatalistaux pfn pred rowdatalist acc =  match rowdatalist with
| [] -> acc
| r::rs -> if (pfn pred r.data) then (filterrowdatalistaux pfn pred rs (acc @ [r])) 
            else (filterrowdatalistaux pfn pred rs acc);;

let filterrowdatalist pred rowdatalist = filterrowdatalistaux evalPred pred rowdatalist [];;

let filtertable pred table = pred table.rowdatalist;;

(* DELETE *)

let negatepred pred = match pred.predoper with
| OpEq -> {pred with predoper=OpNe}
| OpNe -> {pred with predoper=OpEq};;

let deleterowsaux pred table = filterrowdatalist pred table.rowdatalist;;
let deleterow dbms pred tablename = 
  let table = getTable dbms.tablelist tablename in 
  let table = substituteRowDatalist table (deleterowsaux (negatepred pred) table) in
  let newtablelist = substituteTable dbms.tablelist table [] in
  {dbms with tablelist = newtablelist};;

(* UPDATE *)

let rec substitutevalueaux2 coldata values = match values with 
| [] -> coldata
| v::vs -> if (v.datacolname = coldata.datacolname) then
              {coldata with value = v.value}
           else
              substitutevalueaux2 coldata vs;;

let rec substitutevalueaux1 rowdata values  = match rowdata with 
| [] -> []
| rd::rds -> (substitutevalueaux2 rd values) ::substitutevalueaux1 rds values;;

let substitutevalueaux row values  = {row with data = (substitutevalueaux1 row.data values)};;

let substitutevaluespred pred row values  = 
  if (evalPred pred row.data) then
    substitutevalueaux row values
   else
    row;;


let rec updaterowsaux rowdatalist pred values = 
match rowdatalist with
| [] -> []
| r::rs -> (substitutevaluespred pred r values)::updaterowsaux rs pred values;;

let updaterows dbms tablename pred values = 
  let table = getTable dbms.tablelist tablename in 
  let table = substituteRowDatalist table (updaterowsaux table.rowdatalist pred values) in
  let newtablelist = substituteTable dbms.tablelist table [] in
  {dbms with tablelist = newtablelist};;


(* ROW DATA RETRIEVAL *)
let str_of_mltablecoldatavalue v = match v with
| Int i -> string_of_int i
| Char s -> s;;

let getRowColValue coldef rowdatalistdata =  
  let rec getRowColValueAux coldef rowdatalistdata acc = match rowdatalistdata with
  | [] -> acc
  | r::rs -> if (r.datacolname = coldef.defcolname) then r.value else (getRowColValueAux coldef rs acc)
  in
  getRowColValueAux coldef rowdatalistdata (Char "NULL");;


(* DISPLAY FUNCTIONS *)

(* COLUMNS *)
let rec printColDeflist coldeflist  = match coldeflist with
| [] -> print_newline()
| c::cs -> print_string (padStringRight c.defcolname); printColDeflist cs;;


(* ROWS *)
let rec printRowDetail coldeflist rowData = match coldeflist with
| []-> print_newline()
| c::cs ->  print_string ( padStringRight ( str_of_mltablecoldatavalue (getRowColValue c rowData))); 
    printRowDetail cs rowData;;

let rec printRowlist coldeflist rowdatalist = match rowdatalist with
| [] -> ()
| r::rs -> printRowDetail coldeflist r.data; printRowlist coldeflist rs;;

let printRowCount rowdatalist = 
  print_string ( (string_of_int (length rowdatalist)) ^ " rows displayed\n");;

let printRowslist coldeflist rowdatalist = 
  printColDeflist coldeflist; 
  print_string("------------------------------------------------------------"); print_newline();
  printRowlist coldeflist rowdatalist; print_newline();
  printRowCount rowdatalist;;

let printTable dbms tablename = 
  let table = getTable dbms.tablelist tablename in 
  print_newline();
  print_string ("TABLE: " ^ tablename); print_newline();
  printRowslist table.coldeflist table.rowdatalist;
  print_newline();;
  
    
(* DATABASE CREATION *)
let createDB databasename = 
  { dbmsname = databasename; tablelist = [] };;



(* TABLE CREATION *)
let createTable dbms tablename =
  {
    dbms with tablelist = dbms.tablelist @ [{tablename = tablename; coldeflist = []; rowdatalist = [] }]
  };;

let rec dropTableAux tablelist tablename acc = match tablelist with
| [] -> acc
| x::xs -> if (x.tablename = tablename) then 
              acc @ xs
            else
              x::(dropTableAux xs tablename acc);;
              
let dropTable dbms tablename = 
       let newtablelist = dropTableAux dbms.tablelist tablename []
       in 
       {dbms with tablelist = newtablelist};;




(* COLUMN ADDITION TO TABLE *)
let addColumn table coldef = 
    {table with coldeflist = table.coldeflist @ [coldef]};;

let addColumnToTable dbms tablename colDef = 
{
      dbms with tablelist = (substituteTable dbms.tablelist (addColumn (getTable dbms.tablelist tablename) colDef) []) 
};;



(* INSERT OF A DATAROW TO A TABLE *)
let addDataRowColData table rowData =
    {table with rowdatalist = table.rowdatalist @ [rowData]};;

let addDataRowToTable dbms tablename rowData = 
{
      dbms with tablelist = (substituteTable dbms.tablelist (addDataRowColData (getTable dbms.tablelist tablename) rowData) []) 
};;


(* AGGREGATES *)
let rec getIntValue rowdata colname = match rowdata with
| [] -> 0
| rc::rcs -> if (rc.datacolname = colname) then (int_of_coldatavalue rc.value) else getIntValue rcs colname
;;

let printaggrtitle aggr = 
  match aggr with
  | Avg -> print_string ("AVERAGE ")
  | Sum -> print_string ("SUM ");;

let rec aggrsum rowdatalist colname aggr = match rowdatalist with
| [] -> 0
| rd::rds -> (getIntValue rd.data colname) + (aggrsum rds colname aggr)
    ;;

let rec aggrmain rowdatalist colname aggr = match aggr with
  | Avg -> (aggrsum rowdatalist colname aggr) / (length rowdatalist)
  | Sum -> aggrsum rowdatalist colname aggr;;
  
let aggregateaux rowdatalist colname aggr = 
  print_string(">>>>>  ");
  printaggrtitle aggr;
  print_string(colname); print_string(": ");
  print_string(string_of_int(aggrmain rowdatalist colname aggr));
  print_string("   <<<<<");
  print_newline(); print_newline();;


let aggregate dbms tablename colname aggr = 
  let table = getTable dbms.tablelist tablename in 
  aggregateaux table.rowdatalist colname aggr;
  dbms;;
  